# -*- coding: utf-8 -*-
'''
cpenv.api
=========
This module provides the main api for cpenv. Members of the api return models
which can be used to manipulate virtualenvs and modules. Members are available
directly from the cpenv namespace.
'''

import os
import virtualenv
import shutil
from .cache import EnvironmentCache
from .resolver import Resolver
from .utils import unipath
from .models import VirtualEnvironment, Module
from .hooks import run_global_hook
from .deps import Git
from . import utils, defaults


def create(name_or_path=None, config=None):
    '''Create a virtual environment. You can pass either the name of a new
    environment to create in your CPENV_HOME directory OR specify a full path
    to create an environment outisde your CPENV_HOME.

    Create an environment in CPENV_HOME::

        >>> cpenv.create('myenv')

    Create an environment elsewhere::

        >>> cpenv.create('~/custom_location/myenv')

    :param name_or_path: Name or full path of environment
    :param config: Environment configuration including dependencies etc...
    '''

    # Get the real path of the environment
    if utils.is_system_path(name_or_path):
        path = unipath(name_or_path)
    else:
        path = unipath(get_home_path(), name_or_path)

    if os.path.exists(path):
        raise OSError('{} already exists'.format(path))

    env = VirtualEnvironment(path)
    utils.ensure_path_exists(env.path)

    if config:
        if utils.is_git_repo(config):
            Git('').clone(config, env.path)
        else:
            shutil.copy2(config, env.config_path)
    else:
        with open(env.config_path, 'w') as f:
            f.write(defaults.environment_config)

    utils.ensure_path_exists(env.hook_path)
    utils.ensure_path_exists(env.modules_path)

    env.run_hook('precreate')

    virtualenv.create_environment(env.path)
    if not utils.is_home_environment(env.path):
        EnvironmentCache.add(env)
        EnvironmentCache.save()

    try:
        env.update()
    except:
        utils.rmtree(path)
        logger.debug('Failed to update, rolling back...')
        raise
    else:
        env.run_hook('postcreate')

    return env


def remove(name_or_path):
    '''Remove an environment or module

    :param name_or_path: name or path to environment or module
    '''

    r = resolve(name_or_path)
    r.resolved[0].remove()

    EnvironmentCache.discard(r.resolved[0])
    EnvironmentCache.save()


def resolve(*args):
    '''Resolve a list of virtual environment and module names then return
    a :class:`Resolver` instance.'''

    r = Resolver(*args)
    r.resolve()
    return r


def activate(*args):
    '''Activate a virtual environment by name or path. Additional args refer
    to modules residing in the specified environment that you would
    also like to activate.

    Activate an environment::

        >>> cpenv.activate('myenv')

    Activate an environment with some modules::

        >>> cpenv.activate('myenv', 'maya', 'mtoa', 'vray_for_maya')

    :param name_or_path: Name or full path of environment
    :param modules: Additional modules to activate
    :returns: :class:`VirtualEnv` instance of active environment
    '''

    r = resolve(*args)
    r.activate()
    return get_active_env()


def launch(module_name, *args, **kwargs):
    '''Activates and launches a module

    :param module_name: name of module to launch
    '''

    r = resolve(module_name)
    r.activate()
    mod = r.resolved[0]
    mod.launch(*args, **kwargs)


def deactivate():
    '''Deactivates an environment by restoring all env vars to a clean state
    stored prior to activating environments
    '''

    if 'CPENV_ACTIVE' not in os.environ or 'CPENV_CLEAN_ENV' not in os.environ:
        raise EnvironmentError('Can not deactivate environment...')

    utils.restore_env_from_file(os.environ['CPENV_CLEAN_ENV'])


def get_home_path():
    ''':returns: your home path...CPENV_HOME env var OR ~/.cpenv'''

    home = unipath(os.environ.get('CPENV_HOME', '~/.cpenv'))
    home_modules = unipath(home, 'modules')
    if not os.path.exists(home):
        os.makedirs(home)
    if not os.path.exists(home_modules):
        os.makedirs(home_modules)
    return home


def get_module_paths():
    ''':returns: paths in CPENV_MODULES env var and CPENV_HOME/modules'''

    module_paths = []

    cpenv_modules_path = os.environ.get('CPENV_MODULES', None)
    if cpenv_modules_path:
        module_paths.extend(cpenv_modules_path.split(os.pathsep))

    module_paths.append(unipath(get_home_path(), 'modules'))

    return module_paths


def get_active_env():
    ''':returns: the active environment as a :class:`VirtualEnvironment`
    instance or None if one is not active.
    '''

    active = os.environ.get('CPENV_ACTIVE', None)
    if active:
        return VirtualEnvironment(active)


def get_environment(name_or_path):
    '''Get a :class:`VirtualEnvironment` by name or path.'''

    r = resolve(name_or_path)
    return r.resolved[0]


def get_environments():
    '''Returns a list of all known virtual environments as
    :class:`VirtualEnvironment` instances. This includes those in CPENV_HOME
    and any others that are cached(created by the current user or activated
    once by full path.)
    '''

    environments = set()

    cwd = os.getcwd()
    for d in os.listdir(cwd):

        if d == 'environment.yml':
            environments.add(VirtualEnvironment(cwd))
            continue

        path = unipath(cwd, d)
        if utils.is_environment(path):
            environments.add(VirtualEnvironment(path))

    home = get_home_path()
    for d in os.listdir(home):

        path = unipath(home, d)
        if utils.is_environment(path):
            environments.add(VirtualEnvironment(path))

    for env in EnvironmentCache:
        environments.add(env)

    return sorted(list(environments), key=lambda x: x.name)


def get_modules():
    '''Returns a list of available modules.'''

    modules = set()

    cwd = os.getcwd()
    for d in os.listdir(cwd):

        if d == 'module.yml':
            modules.add(Module(cwd))

        path = unipath(cwd, d)
        if utils.is_module(path):
            modules.add(Module(cwd))

    module_paths = get_module_paths()
    for module_path in module_paths:
        for d in os.listdir(module_path):

            path = unipath(module_path, d)
            if utils.is_module(path):
                modules.add(Module(path))

    return sorted(list(modules), key=lambda x: x.name)


def get_active_modules():
    ''':returns: a list of active :class:`Module` s or []'''

    modules = os.environ.get('CPENV_ACTIVE_MODULES', None)
    if modules:
        modules = modules.split(os.pathsep)
        return [Module(module) for module in modules]

    return []


def add_active_module(module):
    '''Add a module to CPENV_ACTIVE_MODULES environment variable'''

    modules = set(get_active_modules())
    modules.add(module)
    new_modules_path = os.pathsep.join([m.path for m in modules])
    os.environ['CPENV_ACTIVE_MODULES'] = str(new_modules_path)


def rem_active_module(module):
    '''Remove a module from CPENV_ACTIVE_MODULES environment variable'''

    modules = set(get_active_modules())
    modules.discard(module)
    new_modules_path = os.pathsep.join([m.path for m in modules])
    os.environ['CPENV_ACTIVE_MODULES'] = str(new_modules_path)


def create_module(name_or_path, config=None, branch=None):

    # Get the real path of the module
    if utils.is_system_path(name_or_path):
        path = unipath(name_or_path)
    else:
        path = unipath('.', name_or_path)

    if os.path.exists(path):
        raise OSError('{} already exists'.format(path))

    utils.ensure_path_exists(path)
    module = Module(path)

    if config:
        if utils.is_git_repo(config):
            Git('').clone(config, module.path, branch)
        elif utils.is_module(config):
            shutil.copytree(unipath(config), module.path)
        elif os.path.isfile(config) and config.endswith('.yml'):
            utils.ensure_path_exists(module.path)
            shutil.copy2(config, module.config_path)
        else:
            raise Exception('Config must be a repo, module, or config_path.')
    else:
        with open(module.config_path, 'w') as f:
            f.write(defaults.module_config)

    utils.ensure_path_exists(module.hook_path)
    module.run_hook('precreatemodule')
    module.run_hook('postcreatemodule')

    return module



environment_config = '''
# Available variables
# $ENVIRON - path to this environment
# $PLATFORM - path to platform
# $PYVER - path to python version

# Wrap variables in brackets when they are nested in valid string Template
# characters. For example:

# DO 'this${variable}isnested/' NOT 'this$variableisnested'

environment: {}
    # RELATIVE_VAR: '$ENVIRON/relative/path'
    # PLATFORM_VAR: '$ENVIRON/$PLATFORM/path'
    # PLATFORM_VARB:
    #    win: 'C:/windows/path'
    #    linux: '/linux/path'
    #    osx: '/mac/path'
dependencies:
    modules: []
        # - name: maya2017
        #   repo: https://github.com/cpenv/maya_module.git
        #   branch: 2017
    git: []
        # - path: '$ENVIRON/path/to/clone'
        #   repo: https://github.com/random/repo.git
        #   branch: master
    pip: []
        # - requests
'''

module_config = '''
# Available variables
# $ENVIRON - path to this modules parent environment
# $MODULE - path to this module
# $PLATFORM - path to platform
# $PYVER - path to python version

# Wrap variables in brackets when they are nested in valid string Template
# characters. For example:

# DO 'this${variable}isnested/' NOT 'this$variableisnested'

environment: {}
    # RELATIVE_VAR: '$ENVIRON/relative/path'
    # PLATFORM_VAR: '$ENVIRON/$PLATFORM/path'
    # PLATFORM_VARB:
    #    win: 'C:/windows/path'
    #    linux: '/linux/path'
    #    osx: '/mac/path'
dependencies:
    modules: []
        # - name: maya2017
        #   repo: https://github.com/cpenv/maya_module.git
        #   branch: 2017
    git: []
        # - path: '$ENVIRON/path/to/clone'
        #   repo: https://github.com/random/repo.git
        #   branch: master
    pip: []
        # - requests
'''


# -*- coding: utf-8 -*-
'''
cpenv.resolver
==============
Defines :class:`Resolver` used to resolve cpenv :class:`VirtualEnvironment` s and :class:`Module` s
'''
import os
from .models import VirtualEnvironment, Module
from .utils import (unipath, is_environment, is_module, join_dicts,
                    set_env, walk_up, is_redirecting, redirect_to_env_paths)
from .cache import EnvironmentCache


class ResolveError(Exception):
    pass


class Resolver(object):
    '''Resolve, combine, activate :class:`VirtualEnvironment` and :class:`Module` environments.
    The args passed can follow two possible signatures.

    usage::

        >>> Resolver('production_env', 'module_a', 'module_b', 'module_c')

    If the first argument passed is a VirtualEnvironment, each additional arg
    is looked up relative to that VirtualEnvironment.

    OR::

        >>> Resolver('module_a', 'module_b', 'module_c')

    In this case the modules are looked up in the active VirtualEnvironment.

    :param cache: Optional keyword argument, cache for use with path resolvers

    '''

    def __init__(self, *args, **kwargs):

        if not args:
            raise ValueError('Resolver expects at least 1 argument.')

        self.paths = args
        self.cache = kwargs.get('cache', EnvironmentCache)
        self.resolved = None
        self.combined = None

    def resolve(self):

        self.resolved = []
        paths = list(self.paths)

        for resolver in resolvers:
            try:

                resolved = resolver(self, paths[0])

                if isinstance(resolved, VirtualEnvironment):
                    self.resolved.append(resolved)
                    paths.pop(0)

                elif isinstance(resolved, list):
                    self.resolved.extend(resolved)
                    paths.pop(0)

                break

            except ResolveError:
                continue

        for path in paths:
            for resolver in module_resolvers:
                try:

                    resolved = resolver(self, path)

                    if isinstance(resolved, Module):
                        self.resolved.append(resolved)

                    elif isinstance(resolved, list):
                        self.resolved.extend(resolved)

                    break

                except ResolveError:
                    continue
            else:
                raise ResolveError(
                    'Could not resolve environment or module: ' + path
                )

        return self.resolved

    def combine(self):

        if not self.resolved:
            raise ValueError('You must call Resolver.resolve first.')

        return join_dicts(*[obj.environment for obj in self.resolved])

    def activate(self):

        set_env(self.combine())

        for obj in self.resolved:
            obj.activate()


def path_is_venv_resolver(resolver, path):
    '''Checks if path is already a VirtualEnvironment'''

    if isinstance(path, VirtualEnvironment):
        return path

    raise ResolveError


def path_resolver(resolver, path):
    '''Resolves VirtualEnvironments with a relative or absolute path'''

    path = unipath(path)

    if is_environment(path):
        return VirtualEnvironment(path)

    raise ResolveError


def home_resolver(resolver, path):
    '''Resolves VirtualEnvironments in CPENV_HOME'''

    from .api import get_home_path

    path = unipath(get_home_path(), path)

    if is_environment(path):
        return VirtualEnvironment(path)

    raise ResolveError


def cache_resolver(resolver, path):
    '''Resolves VirtualEnvironments in EnvironmentCache'''

    env = resolver.cache.find(path)
    if env:
        return env

    raise ResolveError


def path_is_module_resolver(resolver, path):
    '''Checks if path is already a :class:`Module` object'''

    if isinstance(path, Module):
        return path

    raise ResolveError


def module_resolver(resolver, path):
    '''Resolves module in previously resolved environment.'''

    if resolver.resolved:

        if isinstance(resolver.resolved[0], VirtualEnvironment):
            env = resolver.resolved[0]
            mod = env.get_module(path)

            if mod:
                return mod

    raise ResolveError


def modules_path_resolver(resolver, path):
    '''Resolves modules in CPENV_MODULES path and CPENV_HOME/modules'''

    from .api import get_module_paths

    for module_dir in get_module_paths():
        mod_path = unipath(module_dir, path)

        if is_module(mod_path):
            return Module(mod_path)

    raise ResolveError


def active_env_module_resolver(resolver, path):
    '''Resolves modules in currently active environment.'''

    from .api import get_active_env

    env = get_active_env()
    if not env:
        raise ResolveError

    mod = env.get_module(path)
    if not mod:
        raise ResolveError

    return mod


def redirect_resolver(resolver, path):
    '''Resolves environment from .cpenv file...recursively walks up the tree
    in attempt to find a .cpenv file'''

    if not os.path.exists(path):
        raise ResolveError

    if os.path.isfile(path):
        path = os.path.dirname(path)

    for root, _, _ in walk_up(path):
        if is_redirecting(root):
            env_paths = redirect_to_env_paths(unipath(root, '.cpenv'))
            r = Resolver(*env_paths)
            return r.resolve()

    raise ResolveError


resolvers = [
    path_is_venv_resolver,
    path_resolver,
    home_resolver,
    cache_resolver,
    redirect_resolver,
]

module_resolvers = [
    path_is_module_resolver,
    module_resolver,
    active_env_module_resolver,
    modules_path_resolver
]



# -*- coding: utf-8 -*-
import os
from operator import itemgetter
import collections
from functools import partial
import click
import colorama
import cpenv
import cpenv.utils as utils
import cpenv.shell as shell
from cpenv.cache import EnvironmentCache


red = partial(click.style, fg='red')
green = partial(click.style, fg='green')
blue = partial(click.style, fg='blue')
yellow = partial(click.style, fg='yellow')
cyan = partial(click.style, fg='cyan')
bold_red = partial(click.style, fg='red', bold=True)
bold_green = partial(click.style, fg='green', bold=True)
bold_blue = partial(click.style, fg='blue', bold=True)
bold_yellow = partial(click.style, fg='yellow', bold=True)
bold_cyan = partial(click.style, fg='cyan', bold=True)
bold = partial(click.style, bold=True)


def get_type(obj):
    '''Get the type or classification of an environment or module'''
    # TODO This functionality should be moved to object models

    if isinstance(obj, cpenv.VirtualEnvironment):
        return 'python'

    return 'local' if obj.parent else 'shared'


def _type_and_name(obj):
    '''Sort key for a list of environments and modules'''

    return isinstance(obj, cpenv.Module), obj.name


def contains_env(modules):
    return any([isinstance(m, cpenv.VirtualEnvironment) for m in modules])


def get_info(obj, indent=0, root=None):

    name = ' ' * indent + obj.name
    type_ = get_type(obj)
    if root:
        path = '...' + os.path.sep + os.path.relpath(obj.path, root)
    else:
        path = obj.path

    return name, type_, path


def format_objects(objects, children=False, columns=None, header=True):
    '''Format a list of environments and modules for terminal output'''

    columns = columns or ('NAME', 'TYPE', 'PATH')
    objects = sorted(objects, key=_type_and_name)
    data = []
    for obj in objects:
        if isinstance(obj, cpenv.VirtualEnvironment):
            data.append(get_info(obj))
            modules = obj.get_modules()
            if children and modules:
                for mod in modules:
                    data.append(get_info(mod, indent=2, root=obj.path))
        else:
            data.append(get_info(obj))

    maxes = [len(max(col, key=len)) for col in zip(*data)]
    tmpl = '{:%d}  {:%d}  {:%d}' % tuple(maxes)
    lines = []
    if header:
        lines.append('\n' + bold_blue(tmpl.format(*columns)))

    for obj_data in data:
        lines.append(tmpl.format(*obj_data))

    return '\n'.join(lines)


@click.group()
@click.version_option(cpenv.__version__)
def cli():
   '''Cpenv commands'''
   EnvironmentCache.validate()


@cli.command()
def info():
    '''Show context info'''

    env = cpenv.get_active_env()
    modules = []
    if env:
        modules = env.get_modules()
    active_modules = cpenv.get_active_modules()

    if not env and not modules and not active_modules:
        click.echo('\nNo active modules...')
        return

    click.echo(bold('\nActive modules'))
    if env:
        click.echo(format_objects([env] + active_modules))

        available_modules = set(modules) - set(active_modules)
        if available_modules:

            click.echo(
                bold('\nInactive modules in {}\n').format(cyan(env.name))
            )
            click.echo(format_objects(available_modules, header=False))

    else:
        click.echo(format_objects(active_modules))

    available_shared_modules = set(cpenv.get_modules()) - set(active_modules)
    if not available_shared_modules:
        return

    click.echo(bold('\nInactive shared modules \n'))
    click.echo(format_objects(available_shared_modules, header=False))


@cli.command('list')
def list_():
    '''List available environments and modules'''

    environments = cpenv.get_environments()
    modules = cpenv.get_modules()
    click.echo(format_objects(environments + modules, children=True))


@cli.command()
@click.argument('paths', nargs=-1)
@click.option('--skip_local', is_flag=True, help='skip local modules')
@click.option('--skip_shared', is_flag=True, help='skip shared modules')
def activate(paths, skip_local, skip_shared):
    '''Activate an environment'''


    if not paths:
        ctx = click.get_current_context()
        if cpenv.get_active_env():
            ctx.invoke(info)
            return

        click.echo(ctx.get_help())
        examples = (
            '\nExamples: \n'
            '    cpenv activate my_env\n'
            '    cpenv activate ./relative/path/to/my_env\n'
            '    cpenv activate my_env my_module\n'
        )
        click.echo(examples)
        return

    if skip_local:
        cpenv.module_resolvers.remove(cpenv.resolver.module_resolver)
        cpenv.module_resolvers.remove(cpenv.resolver.active_env_module_resolver)

    if skip_shared:
        cpenv.module_resolvers.remove(cpenv.resolver.modules_path_resolver)

    try:
        r = cpenv.resolve(*paths)
    except cpenv.ResolveError as e:
        click.echo('\n' + str(e))
        return

    resolved = set(r.resolved)
    active_modules = set()
    env = cpenv.get_active_env()
    if env:
        active_modules.add(env)
    active_modules.update(cpenv.get_active_modules())

    new_modules = resolved - active_modules
    old_modules = active_modules & resolved

    if old_modules and not new_modules:
        click.echo(
            '\nModules already active: '
            + bold(' '.join([obj.name for obj in old_modules]))
        )
        return

    if env and contains_env(new_modules):
        click.echo('\nUse bold(exit) to leave your active environment first.')
        return

    click.echo('\nResolved the following modules...')
    click.echo(format_objects(r.resolved))
    r.activate()
    click.echo(blue('\nLaunching subshell...'))

    modules = sorted(resolved | active_modules, key=_type_and_name)
    prompt = ':'.join([obj.name for obj in modules])
    shell.launch(prompt)


@cli.command()
@click.argument('name_or_path', required=False)
@click.option('--config', help='Path to config')
def create(name_or_path, config):
    '''Create a new environment.'''

    if not name_or_path:
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        examples = (
            '\nExamples:\n'
            '    cpenv create my_env\n'
            '    cpenv create ./relative/path/to/my_env\n'
            '    cpenv create my_env --config ./relative/path/to/config\n'
            '    cpenv create my_env --config git@github.com:user/config.git\n'
        )
        click.echo(examples)
        return

    click.echo(
        blue('Creating a new virtual environment ' + name_or_path)
    )
    try:
        env = cpenv.create(name_or_path, config)
    except Exception as e:
        click.echo(bold_red('FAILED TO CREATE ENVIRONMENT!'))
        click.echo(e)
    else:
        click.echo(bold_green('Successfully created environment!'))
    click.echo(blue('Launching subshell'))

    cpenv.activate(env)
    shell.launch(env.name)


@cli.command()
@click.argument('name_or_path')
def remove(name_or_path):
    '''Remove an environment'''

    click.echo()
    try:
        r = cpenv.resolve(name_or_path)
    except cpenv.ResolveError as e:
        click.echo(e)
        return

    obj = r.resolved[0]
    if not isinstance(obj, cpenv.VirtualEnvironment):
        click.echo('{} is a module. Use `cpenv module remove` instead.')
        return

    click.echo(format_objects([obj]))
    click.echo()

    user_confirmed = click.confirm(
        red('Are you sure you want to remove this environment?')
    )
    if user_confirmed:
        click.echo('Attempting to remove...', nl=False)

        try:
            obj.remove()
        except Exception as e:
            click.echo(bold_red('FAIL'))
            click.echo(e)
        else:
            click.echo(bold_green('OK!'))


@cli.group()
def cache():
    '''Environment cache commands'''


@cache.command('list')
def list_():
    '''List available environments and modules'''

    click.echo('Cached Environments')

    environments = list(EnvironmentCache)
    click.echo(format_objects(environments, children=False))


@cache.command()
@click.argument('path')
def add(path):
    '''Add an environment to the cache. Allows you to activate the environment
    by name instead of by full path'''

    click.echo('\nAdding {} to cache......'.format(path), nl=False)
    try:
        r = cpenv.resolve(path)
    except Exception as e:
        click.echo(bold_red('FAILED'))
        click.echo(e)
        return

    if isinstance(r.resolved[0], cpenv.VirtualEnvironment):
        EnvironmentCache.add(r.resolved[0])
        EnvironmentCache.save()
        click.echo(bold_green('OK!'))


@cache.command()
@click.argument('path')
def remove(path):
    '''Remove a cached environment. Removed paths will no longer be able to
    be activated by name'''

    r = cpenv.resolve(path)
    if isinstance(r.resolved[0], cpenv.VirtualEnvironment):
        EnvironmentCache.discard(r.resolved[0])
        EnvironmentCache.save()


@cache.command()
def clear():
    '''Clear environment cache'''

    if click.confirm('Clear the environment cache?'):
        cpenv.EnvironmentCache.clear()
        click.echo('Cache cleared.')


@cli.group()
def module():
    '''Module commands'''


@module.command()
@click.argument('name_or_path')
@click.option('--config', help='path to config')
def create(name_or_path, config):
    '''Create a new template module.

    You can also specify a filesystem path like "./modules/new_module"
    '''

    click.echo('Creating module {}...'.format(name_or_path), nl=False)
    try:
        module = cpenv.create_module(name_or_path, config)
    except Exception as e:
        click.echo(bold_red('FAILED'))
        raise
    else:
        click.echo(bold_green('OK!'))
        click.echo('Browse to your new module and make some changes.')
        click.echo("When you're ready add the module to an environment:")
        click.echo('    cpenv module add my_module ./path/to/my_module')
        click.echo('Or track your module on git and add it directly from the repo:')
        click.echo('    cpenv module add my_module git@github.com:user/my_module.git')


@module.command()
@click.argument('name', required=False)
@click.argument('path', required=False)
@click.option('--branch', help='Branch to clone if specifying a repo')
@click.option(
    '--type',
    help='local to active environment or shared',
    type=click.Choice(['local', 'shared']),
    default='local',
    show_default=True)
def add(name, path, branch, type):
    '''Add a module to an environment. PATH can be a git repository path or
    a filesystem path. '''

    if not name and not path:
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        examples = (
            '\nExamples:\n'
            '    cpenv module add my_module ./path/to/my_module\n'
            '    cpenv module add my_module git@github.com:user/my_module.git'
            '    cpenv module add my_module git@github.com:user/my_module.git --branch=master --type=shared'
        )
        click.echo(examples)
        return

    if not name:
        click.echo('Missing required argument: name')
        return

    if not path:
        click.echo('Missing required argument: path')

    env = cpenv.get_active_env()
    if type=='local':
        if not env:
            click.echo('\nActivate an environment to add a local module.\n')
            return

        if click.confirm('\nAdd {} to active env {}?'.format(name, env.name)):
            click.echo('Adding module...', nl=False)
            try:
                env.add_module(name, path, branch)
            except:
                click.echo(bold_red('FAILED'))
                raise
            else:
                click.echo(bold_green('OK!'))

        return

    module_paths = cpenv.get_module_paths()
    click.echo('\nAvailable module paths:\n')
    for i, mod_path in enumerate(module_paths):
        click.echo('    {}. {}'.format(i, mod_path))
    choice = click.prompt(
        'Where do you want to add your module?',
        type=int,
        default=0
    )
    module_root = module_paths[choice]
    module_path = utils.unipath(module_root, name)
    click.echo('Creating module {}...'.format(module_path), nl=False)
    try:
        cpenv.create_module(module_path, path, branch)
    except:
        click.echo(bold_red('FAILED'))
        raise
    else:
        click.echo(bold_green('OK!'))


@module.command()
@click.argument('name')
@click.option('--local', is_flag=True, help='Ensure we remove a local module')
def remove(name, local):
    '''Remove a module named NAME. Will remove the first resolved module named NAME. You can also specify a full path to a module. Use the --local option
    to ensure removal of modules local to the currently active environment.'''

    click.echo()
    if not local: # Use resolver to find module
        try:
            r = cpenv.resolve(name)
        except cpenv.ResolveError as e:
            click.echo(e)
            return

        obj = r.resolved[0]
    else: # Try to find module in active environment
        env = cpenv.get_active_env()
        if not env:
            click.echo('You must activate an env to remove local modules')
            return

        mod = env.get_module(name)
        if not mod:
            click.echo('Failed to resolve module: ' + name)
            return

        obj = mod

    if isinstance(obj, cpenv.VirtualEnvironment):
        click.echo('{} is an environment. Use `cpenv remove` instead.')
        return

    click.echo(format_objects([obj]))
    click.echo()

    user_confirmed = click.confirm(
        red('Are you sure you want to remove this module?')
    )
    if user_confirmed:
        click.echo('Attempting to remove...', nl=False)

        try:
            obj.remove()
        except Exception as e:
            click.echo(bold_red('FAILED'))
            click.echo(e)
        else:
            click.echo(bold_green('OK!'))


@module.command()
@click.argument('name')
def localize(name):
    '''Copy a global module to the active environment.'''

    env = cpenv.get_active_env()
    if not env:
        click.echo('You need to activate an environment first.')
        return

    try:
        r = cpenv.resolve(name)
    except cpenv.ResolveError as e:
        click.echo('\n' + str(e))

    module = r.resolved[0]
    if isinstance(module, cpenv.VirtualEnvironment):
        click.echo('\nCan only localize a module not an environment')
        return

    active_modules = cpenv.get_active_modules()
    if module in active_modules:
        click.echo('\nCan not localize an active module.')
        return

    if module in env.get_modules():
        click.echo('\n{} is already local to {}'.format(module.name, env.name))
        return

    if click.confirm('\nAdd {} to env {}?'.format(module.name, env.name)):
        click.echo('Adding module...', nl=False)
        try:
            module = env.add_module(module.name, module.path)
        except:
            click.echo(bold_red('FAILED'))
            raise
        else:
            click.echo(bold_green('OK!'))

    click.echo('\nActivate the localize module:')
    click.echo('    cpenv activate {} {}'.format(env.name, module.name))
